/* 
 * Copyright (c) 2003 Stefano Barbato 
 *
 * $Id: recover.c,v 1.3 2003/11/03 13:00:32 tat Exp $
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
//"Usage: recover /dev/disk (or disk image file)");

#define VERSION "0.2"

char * g_usage =
"recover "VERSION",  recover JPEG images from disk image files or devices\n\
Copyright (c) 2003 by Stefano Barbato - All rights reserved.\n\
Usage: recover infile\n";

#define die_if(a, msg) do { do_die_if(a, msg, __LINE__); } while(0);
void do_die_if(int b, char* msg, int line)
{
	if(!b)
		return;
	fprintf(stderr, "(%d) %s\n", line, msg);
	if(errno != 0)
		fprintf(stderr, "errno msg: %s\n", strerror(errno));
	exit(1);
}

void save_jpeg(const char* start, const char* end)
{
	die_if(*start != '\xFF', "search 1 error");
	die_if(*(end-1) != '\xD9', "search 2 error");
	printf("Save from %x to %x (size: %d)\n", 
		(unsigned int)start, (unsigned int)end, end-start);
	static int idx = 0;
	char fqn[512];
	int ofd;
	snprintf(fqn, 512, "saved-%.3d.jpg", ++idx);
	printf("Writing %s\n", fqn);
	ofd = open(fqn, O_WRONLY | O_CREAT);
	die_if(ofd < 0, "Error: unable to open output file");
	die_if(write(ofd, start, end-start) < 0, 
			"Error: unable to write to the output file");
	close(ofd);
}

char* search(char* data, int data_sz, char* pattern, int pat_sz)
{
	char *p = data, *end = data + data_sz, *j = pattern;

	while(p < end)
	{
		if(*j == 0)
		{
			j = pattern;
			//printf("found at offset: %x\n", p-data-pat_sz+1);
			return p-pat_sz+1;
		}
		if(*j == '.' || *p == *j)
			++j;
		else
			j = pattern;
		++p;
	}
	return (char*)-1;
}


int main(int argc, char** argv)
{
	die_if(argc < 2, g_usage);
	int fd, sz;
	char *mm, *p, *end;
	struct stat st;
	const char* fqn = argv[1];

	fd = open(fqn, O_RDONLY);
	die_if(fd == 0, "Error: cannot open source file");
	
	die_if(fstat(fd, &st) < 0, "Error: fstat error");

	mm = (char *)mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
	die_if(mm < 0, "Error: unable to mmap file");

	char jpeg_head[] = "\xFF\xD8\xFF...Exif"; // dots are wildcards
	char jpeg_foot[] = "\xFF\xD9"; 
	int jpeg_head_sz = sizeof(jpeg_head), jpeg_foot_sz = sizeof(jpeg_foot);
	p = mm;
	sz = st.st_size;
	while(sz > 0 && (p = search(p, sz, jpeg_head, jpeg_head_sz)) != (char*)-1)
	{
		sz = st.st_size-(p-mm);
		end = search(p, sz, jpeg_foot, jpeg_foot_sz);
		if(end != (char*)-1)
			save_jpeg(p, end+jpeg_foot_sz-1);

		p += jpeg_head_sz;
	}

	munmap(mm, st.st_size);
	close(fd);
	return 0;
}

